"""
@file
@brief Contains functions to import a text file into a database (SQLite).
"""

import os
from .database_main import Database


def import_flatfile_into_database(filedb, filetext, table=None, header=True,
                                  columns=None, engine='SQLite', host='localhost', add_key=None,
                                  encoding="utf-8", fLOG=print):
    """

    Function which imports a file into a database.
    It the table exists, it removes it first. There is no addition.

    @param  filedb      something.db3
    @param  filetext    something.txt or .tsv
    @param  table       table name (in the database), if None, the database name will be the filename without extension
    @param  columns     if header is False, this must be specified. It should be a list of column names.
    @param  header      boolean (does it have a header or not)
    @param  engine      engine to use when using a SQL server (SQLite or ODBCMSSQL)
    @param  host        host (server)
    @param  fLOG        logging function (will display information through the command line)
    @param  add_key     name of a key to add (or None if nothing to add)
    @param  encoding    encoding
    @return             table name

    .. exref::
        :title: Import a flat file into a SQLite database
        :tag: SQL

        ::

            from pyensae import import_flatfile_into_database
            dbf = "database.db3"
            file = "textfile.txt"
            import_flatfile_into_database(dbf, file)

        On Windows, `SQLiteSpy <http://www.yunqa.de/delphi/doku.php/products/sqlitespy/index>`_ is a free tool
        very useful to run SQL queries against a sqlite3 database.
    """
    # connection
    db = Database(filedb, engine=engine, host=host, LOG=fLOG)
    db.connect()

    if table is None:
        table = os.path.splitext(
            os.path.split(filetext)[-1])[0].replace(".", "").replace(",", "")

    if db.has_table(table):
        fLOG("remove ", table)
        db.remove_table(table)

    if header:
        columns = None

    db.import_table_from_flat_file(filetext, table, columns=columns,
                                   header=header, add_key=add_key,
                                   encoding=encoding)

    db.close()
    return table



# Generated from \RFilter.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .RFilter import RFilter
else:
    from RFilter import RFilter

# This class defines a complete listener for a parse tree produced by RFilter.


class RFilterListener(ParseTreeListener):

    # Enter a parse tree produced by RFilter#stream.
    def enterStream(self, ctx: RFilter.StreamContext):
        pass

    # Exit a parse tree produced by RFilter#stream.
    def exitStream(self, ctx: RFilter.StreamContext):
        pass

    # Enter a parse tree produced by RFilter#eat.
    def enterEat(self, ctx: RFilter.EatContext):
        pass

    # Exit a parse tree produced by RFilter#eat.
    def exitEat(self, ctx: RFilter.EatContext):
        pass

    # Enter a parse tree produced by RFilter#elem.
    def enterElem(self, ctx: RFilter.ElemContext):
        pass

    # Exit a parse tree produced by RFilter#elem.
    def exitElem(self, ctx: RFilter.ElemContext):
        pass

    # Enter a parse tree produced by RFilter#atom.
    def enterAtom(self, ctx: RFilter.AtomContext):
        pass

    # Exit a parse tree produced by RFilter#atom.
    def exitAtom(self, ctx: RFilter.AtomContext):
        pass

    # Enter a parse tree produced by RFilter#op.
    def enterOp(self, ctx: RFilter.OpContext):
        pass

    # Exit a parse tree produced by RFilter#op.
    def exitOp(self, ctx: RFilter.OpContext):
        pass


del RFilter



"""
@file
@brief Shortcuts to graph_helper
"""

from .blockdiag_helper import draw_diagram
from .corrplot import Corrplot
from .graphviz_helper import run_dot
from .mapplot import plot_map_france, plot_map_france_polygon



# -*- coding: utf-8 -*-
"""
@file
@brief Various functions about graphviz.

.. versionadded:: 1.1
"""

import os
from pyquickhelper.helpgen.conf_path_tools import find_graphviz_dot
from pyquickhelper.loghelper import run_cmd


def dot_exe():
    """
    retrieve graphivz executable

    @return     filename

    .. versionadded:: 1.1
    """
    r = find_graphviz_dot()
    if r is None:
        raise FileNotFoundError("unable to find graphviz")
    return r


def run_dot(dot_file, outimg):
    """
    calls graphivz on a dot file and produces an image

    @param      dot_file        file, format `DOT <http://www.graphviz.org/doc/info/lang.html>`_
    @param      outimg          output image
    @return                     out, err (stdout, stderr from graphviz)

    .. versionadded:: 1.1
    """
    ext = os.path.splitext(outimg)[-1].strip(".")
    exe = dot_exe()
    cmd = "\"{0}\" -T{1} -o{2} {3}".format(exe, ext, outimg, dot_file)
    out, err = run_cmd(cmd, wait=True)
    if len(err) > 0:
        raise Exception("unable to run graphviz on {0}.\nCMD:\n{1}\nOUT:\n{2}\nERR:\n{3}".format(
            dot_file, cmd, out, err))
    return out, err



"""
@file
@brief Use grammar
This module requires `antlr4-python3-runtime <https://pypi.python.org/pypi/antlr4-python3-runtime/>`_.
"""
import os
from antlr4 import ParseTreeWalker, CommonTokenStream, InputStream
from .tree_string_listener import TreeStringListener
from .tree_graph_listener import TreeGraphListener


def get_parser_lexer(language):
    """
    Returns two classes, a parser and a lexer from :epkg:`antlr4`.

    @param      language    to analyse
    @return                 Parser, Lexer
    """
    try:
        if language == "C#" or language == "CSharp":
            language = "CSharp4"

        if language == "R":
            from .RLexer import RLexer
            from .RParser import RParser
            return RParser, RLexer
        elif language == "CSharp4":
            from .CSharpLexer import CSharpLexer
            from .CSharpParser import CSharpParser
            return CSharpParser, CSharpLexer
        elif language == "SQLite":
            from .SQLiteLexer import SQLiteLexer
            from .SQLiteParser import SQLiteParser
            return SQLiteParser, SQLiteLexer
        elif language == "DOT":
            from .DOTLexer import DOTLexer
            from .DOTParser import DOTParser
            return DOTParser, DOTLexer
        elif language == "Python3":
            from .Python3Lexer import Python3Lexer
            from .Python3Parser import Python3Parser
            return Python3Parser, Python3Lexer
        elif language == "SimpleWorkflow":
            from .SimpleWorkflowLexer import SimpleWorkflowLexer
            from .SimpleWorkflowParser import SimpleWorkflowParser
            return SimpleWorkflowParser, SimpleWorkflowLexer
        else:
            folder = os.path.dirname(__file__)
            if folder == "":
                folder = "."
            files = os.listdir(folder)
            raise ImportError(
                "unable to import parsers for language: " +
                language +
                "\navailable files:\n{0}".format(
                    "\n".join(files)))
    except ImportError as e:
        folder = os.path.dirname(__file__)
        if folder == "":
            folder = "."
        files = os.listdir(folder)
        raise ImportError(
            "unable to import parsers for language: " +
            language +
            "\navailable files:\n{0}".format(
                "\n".join(files))) from e


def parse_code(code, class_parser, class_lexer):
    """
    Parses a code and returns a tree.

    @param      code                code to parse
    @param      class_parser        parser
    @param      class_lexer         lexer
    @return                         parsed code
    """
    if isinstance(code, str):
        # we assume it is a string
        code = InputStream(code)

    lexer = class_lexer(code)
    stream = CommonTokenStream(lexer)
    parser = class_parser(stream)
    return parser


def get_tree_string(tree, parser, format=TreeStringListener):
    """
    Returns a string which shows the parsed tree.

    @param      tree        from @see fn parse_code
    @param      parser      the parser used to build the tree, output of @see fn parse_code
    @param      format      None or a class `ParseTreeListener <https://github.com/antlr/antlr4-python3/blob/master/src/antlr4/tree/Tree.py>`_
    @return                 string
    """
    if format is None:
        return tree.toStringTree()
    else:
        walker = ParseTreeWalker()
        listen = format(parser)
        walker.walk(listen, tree)
        return str(listen)


def get_tree_graph(tree, parser, format=TreeGraphListener):
    """
    Returns a graph with :epkg:`networkx`.

    @param      tree        from @see fn parse_code
    @param      parser      the parser used to build the tree, output of @see fn parse_code
    @param      format      None or a class `ParseTreeListener <https://github.com/antlr/antlr4-python3/blob/master/src/antlr4/tree/Tree.py>`_
    @return                 string
    """
    if format is None:
        raise TypeError(  # pragma: no cover
            "format cannot be None")
    walker = ParseTreeWalker()
    listen = format(parser)
    walker.walk(listen, tree)
    return listen



# Generated from \SimpleWorkflow.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .SimpleWorkflowParser import SimpleWorkflowParser
else:
    from SimpleWorkflowParser import SimpleWorkflowParser

# This class defines a complete listener for a parse tree produced by SimpleWorkflowParser.


class SimpleWorkflowListener(ParseTreeListener):

    # Enter a parse tree produced by SimpleWorkflowParser#parse.
    def enterParse(self, ctx: SimpleWorkflowParser.ParseContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#parse.
    def exitParse(self, ctx: SimpleWorkflowParser.ParseContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#final_stmt.
    def enterFinal_stmt(self, ctx: SimpleWorkflowParser.Final_stmtContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#final_stmt.
    def exitFinal_stmt(self, ctx: SimpleWorkflowParser.Final_stmtContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#affectation_stmt_comma.
    def enterAffectation_stmt_comma(self, ctx: SimpleWorkflowParser.Affectation_stmt_commaContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#affectation_stmt_comma.
    def exitAffectation_stmt_comma(self, ctx: SimpleWorkflowParser.Affectation_stmt_commaContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#affectation_stmt.
    def enterAffectation_stmt(self, ctx: SimpleWorkflowParser.Affectation_stmtContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#affectation_stmt.
    def exitAffectation_stmt(self, ctx: SimpleWorkflowParser.Affectation_stmtContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#for_stmt.
    def enterFor_stmt(self, ctx: SimpleWorkflowParser.For_stmtContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#for_stmt.
    def exitFor_stmt(self, ctx: SimpleWorkflowParser.For_stmtContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#if_stmt.
    def enterIf_stmt(self, ctx: SimpleWorkflowParser.If_stmtContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#if_stmt.
    def exitIf_stmt(self, ctx: SimpleWorkflowParser.If_stmtContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#evaluation_function.
    def enterEvaluation_function(self, ctx: SimpleWorkflowParser.Evaluation_functionContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#evaluation_function.
    def exitEvaluation_function(self, ctx: SimpleWorkflowParser.Evaluation_functionContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#expression.
    def enterExpression(self, ctx: SimpleWorkflowParser.ExpressionContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#expression.
    def exitExpression(self, ctx: SimpleWorkflowParser.ExpressionContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#expression_no_binary.
    def enterExpression_no_binary(self, ctx: SimpleWorkflowParser.Expression_no_binaryContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#expression_no_binary.
    def exitExpression_no_binary(self, ctx: SimpleWorkflowParser.Expression_no_binaryContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#function_call.
    def enterFunction_call(self, ctx: SimpleWorkflowParser.Function_callContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#function_call.
    def exitFunction_call(self, ctx: SimpleWorkflowParser.Function_callContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#variable_name.
    def enterVariable_name(self, ctx: SimpleWorkflowParser.Variable_nameContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#variable_name.
    def exitVariable_name(self, ctx: SimpleWorkflowParser.Variable_nameContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#binary_operator.
    def enterBinary_operator(self, ctx: SimpleWorkflowParser.Binary_operatorContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#binary_operator.
    def exitBinary_operator(self, ctx: SimpleWorkflowParser.Binary_operatorContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#unary_operator.
    def enterUnary_operator(self, ctx: SimpleWorkflowParser.Unary_operatorContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#unary_operator.
    def exitUnary_operator(self, ctx: SimpleWorkflowParser.Unary_operatorContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#stmt_comma.
    def enterStmt_comma(self, ctx: SimpleWorkflowParser.Stmt_commaContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#stmt_comma.
    def exitStmt_comma(self, ctx: SimpleWorkflowParser.Stmt_commaContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#stmt.
    def enterStmt(self, ctx: SimpleWorkflowParser.StmtContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#stmt.
    def exitStmt(self, ctx: SimpleWorkflowParser.StmtContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#connect_stmt.
    def enterConnect_stmt(self, ctx: SimpleWorkflowParser.Connect_stmtContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#connect_stmt.
    def exitConnect_stmt(self, ctx: SimpleWorkflowParser.Connect_stmtContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#data_or_module_output.
    def enterData_or_module_output(self, ctx: SimpleWorkflowParser.Data_or_module_outputContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#data_or_module_output.
    def exitData_or_module_output(self, ctx: SimpleWorkflowParser.Data_or_module_outputContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#module_input.
    def enterModule_input(self, ctx: SimpleWorkflowParser.Module_inputContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#module_input.
    def exitModule_input(self, ctx: SimpleWorkflowParser.Module_inputContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#data_stmt.
    def enterData_stmt(self, ctx: SimpleWorkflowParser.Data_stmtContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#data_stmt.
    def exitData_stmt(self, ctx: SimpleWorkflowParser.Data_stmtContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#module_stmt.
    def enterModule_stmt(self, ctx: SimpleWorkflowParser.Module_stmtContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#module_stmt.
    def exitModule_stmt(self, ctx: SimpleWorkflowParser.Module_stmtContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#module_call.
    def enterModule_call(self, ctx: SimpleWorkflowParser.Module_callContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#module_call.
    def exitModule_call(self, ctx: SimpleWorkflowParser.Module_callContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#element_name.
    def enterElement_name(self, ctx: SimpleWorkflowParser.Element_nameContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#element_name.
    def exitElement_name(self, ctx: SimpleWorkflowParser.Element_nameContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#list_param_affectation.
    def enterList_param_affectation(self, ctx: SimpleWorkflowParser.List_param_affectationContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#list_param_affectation.
    def exitList_param_affectation(self, ctx: SimpleWorkflowParser.List_param_affectationContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#param_affectation.
    def enterParam_affectation(self, ctx: SimpleWorkflowParser.Param_affectationContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#param_affectation.
    def exitParam_affectation(self, ctx: SimpleWorkflowParser.Param_affectationContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#param_name.
    def enterParam_name(self, ctx: SimpleWorkflowParser.Param_nameContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#param_name.
    def exitParam_name(self, ctx: SimpleWorkflowParser.Param_nameContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#inout_name.
    def enterInout_name(self, ctx: SimpleWorkflowParser.Inout_nameContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#inout_name.
    def exitInout_name(self, ctx: SimpleWorkflowParser.Inout_nameContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#module_name.
    def enterModule_name(self, ctx: SimpleWorkflowParser.Module_nameContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#module_name.
    def exitModule_name(self, ctx: SimpleWorkflowParser.Module_nameContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#data_name.
    def enterData_name(self, ctx: SimpleWorkflowParser.Data_nameContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#data_name.
    def exitData_name(self, ctx: SimpleWorkflowParser.Data_nameContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#constant.
    def enterConstant(self, ctx: SimpleWorkflowParser.ConstantContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#constant.
    def exitConstant(self, ctx: SimpleWorkflowParser.ConstantContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#string_literal.
    def enterString_literal(self, ctx: SimpleWorkflowParser.String_literalContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#string_literal.
    def exitString_literal(self, ctx: SimpleWorkflowParser.String_literalContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#integer_number.
    def enterInteger_number(self, ctx: SimpleWorkflowParser.Integer_numberContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#integer_number.
    def exitInteger_number(self, ctx: SimpleWorkflowParser.Integer_numberContext):
        pass

    # Enter a parse tree produced by SimpleWorkflowParser#real_number.
    def enterReal_number(self, ctx: SimpleWorkflowParser.Real_numberContext):
        pass

    # Exit a parse tree produced by SimpleWorkflowParser#real_number.
    def exitReal_number(self, ctx: SimpleWorkflowParser.Real_numberContext):
        pass


del SimpleWorkflowParser



"""
@file

@brief      generic class to access a SQL database
"""

import os


from pyquickhelper.loghelper.flog import run_cmd


def get_list_server():
    """
    @return     a list of the available servers from this machine
    """
    cmd = "sqlcmd -Lc"
    out, err = run_cmd(cmd, wait=True, shell=True, log_error=False)
    line = out.split("\n")
    line = [_.strip("\r\n ") for _ in line]
    line = [_ for _ in line if len(_) > 0]

    # we place all server containing the machine name in the first place
    machine = os.environ.get(
        "COMPUTERNAME", os.environ.get("HOSTNAME", "--------"))
    line = sorted([(-1 if _.startswith(machine) else 0, _) for _ in line])
    line = [_[1] for _ in line]
    return line


def get_list_instance(server):
    """
    @param      server      server name
    @return                 a list of the available instance on a server
    """
    cmd = 'sqlcmd -S %s -Q "SELECT @@ServerName"' % server
    out, err = run_cmd(cmd, wait=True, shell=True, log_error=False)
    if len(err) > 0:
        cmd = 'sqlcmd -S %s\\SQLEXPRESS -Q "SELECT @@ServerName"' % server
        out, err = run_cmd(cmd, wait=True, shell=True, log_error=False)
        if len(err) > 0:
            raise Exception(
                "unable to find instances for server %s (%s)" %
                (server, err))

    li = [_.strip("\r\n ") for _ in out.split("\n")]
    li = [_ for _ in li if len(_) > 0]
    li = li[1:-1]
    return li


def get_list_database(instance):
    """
    @param      instance    instance name
    @return                 a list of the available database
    """
    cmd = 'sqlcmd -S %s -Q "SELECT name FROM master..sysdatabases"' % instance
    out, err = run_cmd(cmd, wait=True, shell=True, log_error=False)
    li = [_.strip("\r\n ") for _ in out.split("\n")]
    no = ["master", "tempdb", "model", "msdb"]
    li = [_ for _ in li if len(_) > 0 and _ not in no]
    li = li[2:-1]
    return li


def create_database(instance, database, exc=True):
    """
    create a database
    @param      instance    instance name
    @param      database    database name
    @param      exc         if True and if the database exists, raise an Exception

    Example:
    @code
    fLOG (OutputPrint = True)
    fLOG (get_list_database ('PCXAVIER\\SQLEXPRESS'))
    create_database ('PCXAVIER\\SQLEXPRESS', "essai_database")
    drop_database ('PCXAVIER\\SQLEXPRESS', "essai_database")
    @endcode
    """
    ins = get_list_database(instance)
    if database in ins:
        if exc:
            raise Exception("database %s already exists" % database)
    else:
        cmd = 'sqlcmd -S %s -Q "CREATE DATABASE %s"' % (instance, database)
        out, err = run_cmd(cmd, wait=True, shell=True, log_error=False)
        if len(err) > 0:
            raise Exception("error: %s" % err)


def drop_database(instance, database, exc=True):
    """
    remove a database
    @param      instance    instance name
    @param      database    database name
    @param      exc         if True and if the database does not exist, raise an Exception
    """
    ins = get_list_database(instance)
    if database not in ins:
        if exc:
            raise Exception("database %s does not exist" % database)
    else:
        cmd = 'sqlcmd -S %s -Q "DROP DATABASE %s"' % (instance, database)
        out, err = run_cmd(cmd, wait=True, shell=True, log_error=False)
        if len(err) > 0:
            raise Exception("error: %s" % err)



"""
@file
@brief Convert R into Python
"""
import sys
from antlr4 import ParseTreeListener, ParseTreeWalker
from pyquickhelper.pycode import remove_extra_spaces_and_pep8
from .RParser import RParser
from .RLexer import RLexer
from .antlr_grammar_use import parse_code


class R2PyConversionError(Exception):
    """
    Raised when conversion cannot be done.
    """

    def __init__(self, node, message, sofar, sostack):
        """
        """
        try:
            text = node.symbol.text
        except AttributeError:
            text = "ERROR"
        mes = "Unable to convert\n'{0}'\n{1}\n{2}\nPARENT\n'{3}'\n{4}\n---SOFAR---\n{5}\n---SOSTACK---\n{6}\n---END---".format(
            text, type(node), node, ("" if isinstance(
                node, str) else node.parentCtx),
            (str if isinstance(node, str) else type(node.parentCtx)), sofar, sostack)
        mes += "\n------------\nMESSAGE=\n" + str(message)
        Exception.__init__(self, mes)


class TreeStringListener(ParseTreeListener):
    """
    This class is an attempt to run through the tree and convert it into
    a string.
    """

    def __init__(self, parser, fLOG=None):
        """
        constructor

        @param      parser      parser used to parse the code
        @param      fLOG        logging function
        """
        ParseTreeListener.__init__(self)
        self.buffer = []
        self.level = 0
        self.parser = parser
        self.stack = []
        self.elements = []
        self.indent = 0
        self.block = []
        self.memo = []
        self.imports = set()
        self.in_formula = False
        self._fLOG = fLOG
        self.indent_level = {}
        self.add_lambda = False

    def fLOG(self, *l, **p):
        """
        logging
        """
        if self._fLOG:
            self._fLOG(*l, **p)

    def get_python(self):
        """
        Get the Python code for the R code.

        @return     Python code
        """
        def modify(s):
            if s.startswith("from"):
                return s
            else:
                return "import {0}".format(s)
        imports = "\n".join(modify(i) for i in sorted(self.imports))
        if len(imports) > 0:
            imports += "\n\n"
        return imports + "".join(self.elements)

    def add_code(self, node):
        """
        Converts one node into :epkg:`python`.
        """
        name = self.terminal_node_name(node)
        text = node.symbol.text

        if self.in_formula:
            if text in (",", "\n"):
                self.stack.append(("Formop", '"'))
                self.in_formula = False
            self.fLOG("[TreeStringListener]", len(self.block),
                      "~" if self.in_formula else " ", name, text)
        else:
            self.fLOG("[TreeStringListener]", len(self.block), " ", name, text)

        if name == "Parse":
            if text.startswith("#"):
                # Comment
                self.elements.append(text.strip("\n"))
                self.elements.append("\n")
                return self.add_code_final()
            elif text in (";", "\n"):
                # End of an instruction
                self.empty_stack()
                if len(self.elements) > 0 and self.elements[-1] != '\n':
                    self.elements.append("\n")
                return self.add_code_final()
            elif text == "<EOF>":
                return self.add_code_final()
            elif len(text.strip()) == 0:
                return self.add_code_final()
        elif name == "Identifier":
            name_parent = self.terminal_node_name(node.parentCtx)
            if name_parent == "Formula_simple_A":
                if self.in_formula:
                    raise R2PyConversionError(node, name, "".join(
                        self.elements), "\n".join(str(_) for _ in self.stack))
                self.stack.append(("Formop", '"'))
                self.in_formula = True
            elif name_parent == "Functioncall":
                if text == "system.time":
                    self.add_lambda = True
                else:
                    self.add_lambda = False
                self.stack.append(("Functioncall", ""))
            if text == "list":
                self.imports.add("from python2r_helper import list_or_dict")
                self.stack.append((name, "list_or_dict"))
            elif text == "is":
                self.imports.add("from python2r_helper import is_")
                self.stack.append((name, "is_"))
            elif text == "(" and self.add_lambda:
                self.stack.append((name, "lambda: "))
                self.stack.append((name, node))
            else:
                self.stack.append((name, node))
            return self.add_code_final()
        elif name in ("Affectop", "Comparison"):
            self.stack.append((name, node))
            return self.add_code_final()
        elif name == "Constant":
            self.stack.append((name, node))
            return self.add_code_final()
        elif name == "Boolean":
            self.stack.append((name, node))
            return self.add_code_final()
        elif name in ("Functiondef", "Functiondefargs", "Functiondefbody"):
            if text == "{":
                self.empty_stack()
                if len(self.block) == 0:
                    raise R2PyConversionError(node, name, "".join(
                        self.elements), "\n".join(str(_) for _ in self.stack))
                self.block[-1] = True
                return self.add_code_final()
            elif text == "}":
                self.empty_stack()
                if len(self.block) == 0:
                    raise R2PyConversionError(node, name, "".join(
                        self.elements), "\n".join(str(_) for _ in self.stack))
                self.block.pop()
                self.indent -= 1
                self.fLOG(
                    "[TreeStringListener.add_code*] indent -= 1", "--", self.indent)
                return self.add_code_final()
            elif text == '\n':
                return self.add_code_final()
            elif text in ('(', ')', 'function'):
                self.stack.append((name, node))
                return self.add_code_final()
        elif name == "Functiondefargslambda":
            if text == "(":
                return self.add_code_final()
            elif text == ")":
                self.stack.append((name, ":"))
                return self.add_code_final()
        elif name == "Expr":
            if text.startswith("#"):
                # Comment
                self.elements.append("    " * self.indent)
                self.elements.append(text.strip("\n"))
                self.elements.append("\n")
                return self.add_code_final()
            if text in ('(', ')', '[', ']', "+", "-", "|", "&", "||", "&&", "[[", "]]"):
                self.stack.append((None, node))
                return self.add_code_final()
            if text == "!":
                self.stack.append(("Not!", node))
                return self.add_code_final()
            if text == "%*%":
                if sys.version_info[:2] >= (3, 6):
                    self.stack.append((name, "@"))
                else:
                    self.stack.append((name, "*"))
                return self.add_code_final()
            if text in ("%>%", "%<%", "%+%", "%-%", "%/%"):
                self.stack.append((name, text.replace("%", "")))
                return self.add_code_final()
            if text == "\n":
                self.empty_stack()
                if len(self.elements) > 0 and self.elements[-1] != '\n':
                    self.elements.append("\n")
                return self.add_code_final()
            if text == "break":
                self.stack.append((name, node))
                return self.add_code_final()
            if text == "{":
                self.empty_stack()
                if len(self.block) == 0:
                    # We are in an expression.
                    self.stack.append((name, "("))
                    return self.add_code_final()
                else:
                    self.block[-1] = True
                    return self.add_code_final()
            if text == "}":
                # We are in an expression.
                self.empty_stack()
                if len(self.block) == 0:
                    self.stack.append((name, ")"))
                    return self.add_code_final()
                else:
                    self.block.pop()
                    self.indent -= 1
                    self.fLOG(
                        "[TreeStringListener.add_code] indent -= 1", "--", self.indent)
                    return self.add_code_final()
        elif name == "Form":
            if text == "...":
                self.stack.append((None, "*args"))
                return self.add_code_final()
            else:
                self.stack.append((None, node))
                return self.add_code_final()
        elif name == "Argumentname":
            self.stack.append((name, node))
            return self.add_code_final()
        elif name == "Formlist":
            self.stack.append((None, node))
            return self.add_code_final()
        elif name == "Functioncall":
            if text == ')':
                self.stack.append((name, node))
                return self.add_code_final()
            if text == '(':
                self.stack.append((name, node))
                if self.add_lambda:
                    self.stack.append((name, "lambda: "))
                return self.add_code_final()
            if text == '\n':
                return self.add_code_final()
        elif name == "Affectation":
            if text == '\n':
                return self.add_code_final()
        elif name in ("Sub", "Subnobracket"):
            if text == '=':
                # Named parameter.
                self.stack.append((name, node))
                return self.add_code_final()
            if text == ':':
                self.stack.append((name, node))
                return self.add_code_final()
            if text == '...':
                self.stack.append((name, "*args"))
                return self.add_code_final()
        elif name == "Sublist":
            if text == ",":
                self.stack.append((name, node))
                return self.add_code_final()
            if text == '\n':
                return self.add_code_final()
        elif name in ("Exprlist", "Rightexpr"):
            if text.startswith("#"):
                # Comment
                self.empty_stack()
                self.elements.append("  ")
                self.elements.append(text.strip("\n"))
                self.elements.append("\n")
                return self.add_code_final()
            if text in (";", "\n"):
                if self.search_parents(node, "Inlinefunction"):
                    self.empty_stack()
                    self.stack.append(("Inlinefunction", text))
                    return self.add_code_final()
                else:
                    self.empty_stack()
                    if len(self.elements) > 0 and self.elements[-1] != '\n':
                        self.elements.append("\n")
                    return self.add_code_final()
        elif name == "Elseif":
            if text == "if":
                if id(node.parentCtx.parentCtx) not in self.indent_level:
                    raise R2PyConversionError(node, name, "".join(
                        self.elements), "\n".join(str(_) for _ in self.stack))
                else:
                    if self.indent_level[id(node.parentCtx.parentCtx)] != self.indent:
                        raise R2PyConversionError(node, name, "".join(
                            self.elements), "\n".join(str(_) for _ in self.stack))
                self.stack.append((name, "elif"))
                return self.add_code_final()
            elif text == "else":
                # We do nothing. If follows.
                return self.add_code_final()
            else:
                # There should be nothing else.
                pass
        elif name == "Ifelseexpr" or name == "Ifexpr":
            if self.search_parents(node, "Sublist") or self.search_parents(node, "Affectation", 2):
                if text == "if":
                    self.stack.append((name, node))
                    return self.add_code_final()
                elif text == "else":
                    self.stack.append((name, node))
                    return self.add_code_final()
                elif text in ("(", ")"):
                    self.stack.append((name, node))
                    return self.add_code_final()
                elif text == "\n":
                    raise R2PyConversionError(node, name, "".join(
                        self.elements), "\n".join(str(_) for _ in self.stack))
            else:

                if text == "if":
                    self.indent_level[id(node.parentCtx)] = self.indent
                    self.stack.append((name, node))
                    return self.add_code_final()
                elif text == "else":
                    if id(node.parentCtx) not in self.indent_level:
                        raise R2PyConversionError(node, name, "".join(
                            self.elements), "\n".join(str(_) for _ in self.stack))
                    else:
                        if self.indent_level[id(node.parentCtx)] != self.indent:
                            raise R2PyConversionError(node, name, "".join(
                                self.elements), "\n".join(str(_) for _ in self.stack))
                    if len(self.stack):
                        # It does follow }
                        self.empty_stack()
                    if len(self.elements) > 0 and self.elements[-1] != '\n':
                        self.elements.append("\n")

                    self.stack.append((name, node))
                    self.stack.append((":EOL", None))
                    return self.add_code_final()
                elif text == ")":
                    self.stack.append((":EOL", None))
                    return self.add_code_final()
                elif text in ('(', "\n"):
                    return self.add_code_final()
        elif name == "Forexpr":
            if text == "for":
                self.stack.append((name, node))
                return self.add_code_final()
            elif text in ('(', ')'):
                return self.add_code_final()
            elif text == "in":
                self.stack.append((name, node))
                return self.add_code_final()
        elif name == "Whileexpr":
            if text == "while":
                self.stack.append((name, node))
                return self.add_code_final()
            elif text in ('(', ')'):
                return self.add_code_final()
        elif name == "Rangeop":
            if text == ":":
                self.memo.append("range")
                self.stack.append((name, node))
                return self.add_code_final()
            elif text == ":::":
                self.stack.append(("Dotop_static", node))
                return self.add_code_final()
        elif name == "Returnexpr":
            if text == "return":
                self.stack.append((name, node))
                return self.add_code_final()
            if text in ('(', ')'):
                return self.add_code_final()
        elif name == "Formop":
            if text == "~":
                self.imports.add("patsy")
                self.stack.append((name, node))
                return self.add_code_final()
        elif name == "Sublistadd":
            if text == "+":
                self.stack.append((name, node))
                return self.add_code_final()
            elif text == "\n":
                return self.add_code_final()
        elif name == "Dotop":
            if text in ("$", "@"):
                self.stack.append((name, node))
                return self.add_code_final()
        elif name == "Formula_simple_A":
            if text == ".":
                self.stack.append((None, "."))
                return self.add_code_final()
        elif name == "Formula_simple_C":
            if text in ("(", ")", "~"):
                self.stack.append((name, node))
                return self.add_code_final()
        elif name == "Formula_simple_B":
            if text == "within":
                self.imports.add("from python2r_helper import within")
                self.stack.append((name, node))
                return self.add_code_final()
            elif text in ("(",):
                self.stack.append((name, node))
                return self.add_code_final()
            elif text == "{":
                return self.add_code_final()
            elif text == "}":
                return self.add_code_final()
            elif text == ",":
                self.stack.append((name, node))
                self.stack.append(("Formop", '"'))
                return self.add_code_final()
            elif text == ";":
                self.stack.append(("Formop", '"'))
                self.stack.append((name, ","))
                self.stack.append(("Formop", '"'))
                return self.add_code_final()
            elif text == ")":
                self.stack.append(("Formop", '"'))
                self.stack.append((name, node))
                return self.add_code_final()
        elif name == "Operator":
            if text == "%%":
                self.stack.append((name, "%"))
                return self.add_code_final()
            elif text == "^":
                self.stack.append((name, "**"))
                return self.add_code_final()
            else:
                self.stack.append((name, node))
                return self.add_code_final()
        elif name in ("Range_simple", "Range_complexe"):
            if text == ":":
                self.stack.append((name, ","))
                return self.add_code_final()
            else:
                self.stack.append((name, node))
                return self.add_code_final()
        elif name in ("Intersection_simple", "Intersection_complexe"):
            if text == "%in%":
                self.stack.append((name, node))
                return self.add_code_final()
            else:
                self.stack.append((name, node))
                return self.add_code_final()
        elif name == "Implicit_column_name":
            if text == ".":
                self.imports.add("from python2r_helper import ImplicitColumn")
                self.stack.append(("Identifier", "ImplicitColumn"))
                self.stack.append((name, node))
                return self.add_code_final()
            elif text in ("(", ")"):
                return self.add_code_final()
        elif name == "Inlinefunction":
            if text == "{":
                self.inlinefunction = []
                self.stack.append((name, "compile_inline_function('''\n"))
                return self.add_code_final()
            elif text in ("\n", ";"):
                if hasattr(self, "inlinefunction") and len(self.inlinefunction) > 0:
                    self.stack.append((name, text))
                return self.add_code_final()
            elif text == "}":
                self.stack.append((name, "''')"))
                self.inlinefunction = []
                return self.add_code_final()
            else:
                self.inlinefunction.append((name, node))
                return self.add_code_final()

            self.stack.append((name, node))
            return self.add_code_final()

        if text.startswith("#"):
            # Comment
            self.elements.append("    " * self.indent)
            self.elements.append(text.strip("\n"))
            self.elements.append("\n")
            return self.add_code_final()

        raise R2PyConversionError(node, name, "".join(
            self.elements), "\n".join(str(_) for _ in self.stack))

    def add_code_final(self):
        """
        Adds extra characters if needed.
        """
        pass

    def empty_stack(self):
        """
        Empties the stack.
        """
        if len(self.stack) > 0 and len(self.block) > 0 and not self.block[-1]:
            self.indent -= 1
            self.fLOG("[TreeStringListener.empty_stack] indent -= 1",
                      "--", self.indent)
            self.block.pop()

        is_function_def = False
        is_for = False
        as_namespace = False
        for ipos, (name, node) in enumerate(self.stack):
            if name == "Functiondef" and not self.search_parents(node, "Functiondefargslambda", 3):
                is_function_def = True
                break
            elif name == "Forexpr":
                is_for = True
                break
            elif name == "Whileexpr":
                is_for = True
                break
            elif name == "Identifier":
                text = node if isinstance(node, str) else node.symbol.text
                if text == "asNamespace":
                    as_namespace = True
                    break

        if self.indent > 0:
            self.elements.append("    " * self.indent)

        if is_function_def:
            self.elements.append("\n")
            function_name = self.stack[0][1] if isinstance(
                self.stack[0][1], str) else self.stack[0][1].symbol.text
            self.fLOG(
                "[TreeStringListener.empty_stack] add function '{0}'".format(function_name))
            self.elements.append("    " * self.indent)
            self.elements.append("def")
            self.elements.append(" ")
            function_name = function_name.strip(
                '"').replace('.', "_").replace("-", "_")
            self.elements.append(function_name)
            self.stack = self.stack[3:]
            last = self.stack[-1][1].symbol.text
            if last != ")":
                raise R2PyConversionError(self.stack[-1][1], last, "".join(
                    self.elements), "\n".join(str(_) for _ in self.stack))

        # We store some end character we need to add.
        closure = {}

        for ipos, (name, node) in enumerate(self.stack):

            if name == "Functioncall" and node == "":
                # Silent addition.
                continue

            self.fLOG(
                "    cl={0} n={1} - {2}".format(len(closure), name, node))

            if len(closure) > 0 and not isinstance(node, str):
                rem = []
                for k, (leave_node, symbol) in closure.items():
                    b = self.has_parent(node, leave_node)
                    if not b:
                        rem.append(k)
                        args = [symbol, self.terminal_node_name(leave_node),
                                self.terminal_node_name(node), node.symbol.text]
                        self.fLOG(
                            "      closure '{0}' - L-{1} C-{2} ({3})".format(*args))
                        self.elements.append(symbol)
                for r in rem:
                    del closure[r]

            converted = self.to_python(name, node)

            if name == "Identifier":
                converted = converted.replace(".", "_")
            elif name in ("Ifexpr", "Ifelseexpr"):
                if node.symbol.text in ("(", ")") and (self.search_parents(node, "Sublist") or
                                                       self.search_parents(node, "Affectation", 2)):
                    self.fLOG(
                        "      inlineif {0} - '{1}'".format(name, node.symbol.text))
                    if node.symbol.text == "(":
                        converted = "("
                    else:
                        # We need to add ) when leaving this node.
                        converted = ","
                        closure[id(node.parentCtx)] = (node.parentCtx, ")")
            elif name == "Affectop" and self.stack[0][0] in ("Functioncall", "Inlinefunction") and \
                    self.stack[0][1] == "":
                # How to deal with syntax names(df) = something.
                # We add set add a bracket at the end.
                converted = ".set("
                closure[id(node.parentCtx.parentCtx)] = (
                    node.parentCtx.parentCtx, ")")
            elif name == "Formula_simple_C":
                if converted == "(":
                    self.elements.append('"')
                    closure[id(node.parentCtx)] = (node.parentCtx, '"')

            if as_namespace and converted == "=":
                converted = "."

            if self.indent > 0 and converted and self.elements[-1] == "\n":
                self.elements.append("    " * self.indent)

            self.elements.append(converted)
            self.elements.append(" ")
            if is_for and len(self.memo) > 0 and converted == "in":
                self.elements.append(self.memo[-1])
                self.elements.append('(')
            if name == ":EOL":
                self.elements.append(":")
                self.elements.append("\n")
                self.indent += 1
                self.fLOG(
                    "[TreeStringListener.empty_stack-1] indent += 1", "--", self.indent)
                self.block.append(False)

        if len(closure) > 0:
            for k, (leave_node, symbol) in closure.items():
                self.elements.append(symbol)
            # closure = {}

        if is_function_def or is_for:
            if is_for and len(self.memo) > 0:
                self.elements.append(")")
            self.elements.append(":")
            self.elements.append("\n")
            self.indent += 1
            self.fLOG(
                "[TreeStringListener.empty_stack-2] indent += 1", "--", self.indent)
            self.block.append(False)

        self.stack.clear()
        self.memo.clear()

    def search_parents(self, node, substring, max_depth=None):
        """
        Searches for a substring in parents' node name.

        @param      node        current node
        @param      substring   substring to search
        @param      max_depth   number of parents to look at
        @return                 boolean
        """
        if isinstance(node, str):
            return False
        depth = max_depth if max_depth else 0
        n = node.parentCtx
        while (max_depth is None or depth > 0) and n is not None:
            na = self.terminal_node_name(n)
            if substring in na:
                return True
            n = n.parentCtx
            depth -= 1
        return False

    def has_parent(self, current, parent, depth=None):
        """
        Tells if *parent* is one of the parents of *current*.

        @param      current     current node
        @param      parent      parent to look for
        @return                 boolean
        """
        if isinstance(current, str):
            raise NotImplementedError()
        if depth is None:
            ide = 0
        else:
            ide = depth
        n = current
        while (depth is None or ide > 0) and n is not None:
            if id(n) == id(parent):
                return True
            n = n.parentCtx
            ide -= 1
        return False

    def to_python(self, name, node):
        """
        Converts a couple *(name, node)* into :epkg:`Python`.
        """
        if name == "Affectop":
            return "="
        elif name == "Not":
            return " not "
        elif name == "Not!":
            return " ~ "
        elif name == "Boolean":
            text = node.symbol.text
            return text[0] + text[1:].lower()
        elif name == "Dotop":
            return "."
        elif name == "Dotop_static":
            return ".static."
        elif name == "Constant":
            text = node.symbol.text
            if text.startswith("`") and text.endswith("`") and len(text) > 1:
                return 'RCOL("{0}")'.format(text[1:-1])
            if text == "NULL":
                return "None"
            if text.endswith("L"):
                # Integer
                return text[:-1]
            is_formula = self.search_parents(node, "Formula")
            if is_formula and text[0] == '"' and text[-1] == '"':
                return '\\"{0}\\"'.format(text[1:-1])
            else:
                return text
        elif name == "Rangeop":
            text = node.symbol.text
            if text == ":":
                return ","
            else:
                return text
        elif name in ("Ifexpr", "Ifelseexpr"):
            text = node if isinstance(node, str) else node.symbol.text
            if text in ("if", "else") and (self.search_parents(node, "Sublist") or
                                           self.search_parents(node, "Affectation", 2)):
                if text == "if":
                    self.imports.add("from python2r_helper import inlineif")
                    return "inlineif"
                else:
                    return ","
            else:
                return text
        elif name in ("Intersection_simple", "Intersection_complexe"):
            text = node.symbol.text
            if text == "%in%":
                return ") & set("
            else:
                return text
        elif isinstance(node, str):
            return node
        elif node is None:
            return ""
        elif name == "Functiondef":
            if self.search_parents(node, "Functiondeflambda", 2):
                return "lambda"
            else:
                return node.symbol.text
        elif name == "Argumentname":
            text = node.symbol.text
            if text == "lambda":
                return "lambda_"
            else:
                return text.replace(".", "_")
        elif name == "Identifier":
            text = node.symbol.text
            if text == "c":
                # This is a tuple.
                self.imports.add("from python2r_helper import make_tuple")
                return "make_tuple"
            elif text == "class":
                parent_name = self.terminal_node_name(node.parentCtx)
                if parent_name in ("Functioncall", "Subnobracket"):
                    self.imports.add("from python2r_helper import make_class")
                    return "make_class"
                else:
                    return text.replace(".", "_")
            elif text == "finally":
                self.imports.add("from python2r_helper import finallyR")
                return "finallyR"
            elif text == "lambda":
                return "lambda_"
            elif text in ("bquote", "ImplicitColumn"):
                self.imports.add(
                    "from python2r_helper import {0}".format(text))
                return text
            elif text == "try":
                parent_name = self.terminal_node_name(node.parentCtx)
                if parent_name == "Functioncall":
                    self.imports.add("from python2r_helper import dotry")
                    return "dotry"
                else:
                    return text.replace(".", "_")
            else:
                return text.replace(".", "_")
        else:
            text = node.symbol.text
            if text == "c":
                # This is a tuple.
                self.imports.add("from python2r_helper import make_tuple")
                return "make_tuple"
            elif text == "&&":
                return "and"
            elif text == "||":
                return "or"
            else:
                return text

    def terminal_node_name(self, node):
        """
        Converts a terminal node into a rule name.
        """
        return str(type(node.parentCtx)).split('.')[-1].strip("]['><").replace("Context", "")

    def visitTerminal(self, node):
        """
        event
        """
        # node: ['symbol', 'parentCtx']
        # node.symbol: ['source', 'type', 'channel', 'start', 'stop', 'tokenIndex', 'line', 'column', '_text']
        # help(node.parentCtx)
        full_text = node.parentCtx.getText().replace("\n", " EOL ")
        text = node.symbol.text.replace("\n", " EOL ")
        stype = self.terminal_node_name(node)
        text = "{0} [{1} - {2}] {3}".format("    " *
                                            self.level, text, stype, full_text)
        self.buffer.append(text)
        self.add_code(node)

    def visitErrorNode(self, node):
        """
        event
        """
        text = ("    " * self.level) + "error: " + str(node)
        self.buffer.append(text)

    def enterEveryRule(self, ctx):
        """
        event
        """
        kind = str(type(ctx)).split(
            ".")[-1].strip("'<>").replace("Context", "")
        text = ("    " * self.level) + "+ {0}".format(kind)
        self.buffer.append(text)
        self.level += 1

    def exitEveryRule(self, ctx):
        """
        event
        """
        self.level -= 1
        text = ("    " * self.level) + "- "
        self.buffer.append(text)

    def __str__(self):
        """
        usual
        """
        return self.get_python() + "\n----\n" + "\n".join(self.buffer)

    def enterRanges(self, ctx):
        """
        event
        """
        self.fLOG("    add 'range('")
        self.stack.append(("Ranges", "range("))

    def exitRanges(self, ctx):
        """
        event
        """
        self.fLOG("    add ') # range'")
        self.stack.append(("Ranges", ")"))

    def enterIntersections(self, ctx):
        """
        event
        """
        self.fLOG("    add 'set('")
        self.stack.append(("Intersections", "set("))

    def exitIntersections(self, ctx):
        """
        event
        """
        self.fLOG("    add ') # set'")
        self.stack.append(("Intersections", ")"))



"""
@file
@brief Helpers around language grammar.
This module requires `antlr4 <https://pypi.python.org/pypi/antlr4-python3-runtime/>`_.
"""
from antlr4 import ParseTreeListener


class TreeGraphListener(ParseTreeListener):

    """
    This class is an attempt to run through the tree
    and to convert into a graph.

    .. exref::
        :title: Draw a grammar graph for a small code

        ::

            from pyensae.languages import get_parser_lexer, parse_code, get_tree_graph
            from pyensae.graph_helper import run_dot

            code = '''
            namespace hello
            {
                public static class world
                {
                    public static double function(double x, doubly y)
                    {
                        return x+y ;
                    }
                }
            }
            '''

            clparser, cllexer = get_parser_lexer("C#")
            parser = parse_code(code, clparser, cllexer)
            tree = parser.compilation_unit()
            st = get_tree_graph(tree, parser)
            dot = st.to_dot()

            with open(name, "w") as f:
                f.write(dot)
            img = os.path.join(temp, "graph.png")
            run_dot(name, img)
    """

    def __init__(self, parser, verbose=False, fLOG=None):
        """
        constructor

        @param      parser      parser used to parse the code
        @param      verbose     display information along the path
        @param      fLOG        logging function
        """
        ParseTreeListener.__init__(self)
        self.parser = parser
        self.vertices = {}
        self.edges = {}
        self.verbose = verbose
        self.fLOG = fLOG

    @property
    def Vertices(self):
        """
        return vertices
        """
        return self.vertices

    @property
    def Edges(self):
        """
        return edges
        """
        return self.edges

    def _get_key_node(self, node):
        line, col = node.symbol.line, node.symbol.column
        return line, col, "#END#"

    def _get_key_context(self, ctx):
        line, col = ctx.start.line, ctx.start.column
        class_name = ctx.__class__.__name__
        if class_name.endswith("Context"):
            class_name = class_name[0].lower() + class_name[1:-7]
        return line, col, class_name

    def visitTerminal(self, node):
        """
        event
        """
        key = self._get_key_node(node)
        val = (node.getText(), node, node.parentCtx)
        self.vertices[key] = val
        kc = self._get_key_context(node.parentCtx)
        self.edges[kc, key] = 1

    def visitErrorNode(self, node):
        """
        event
        """
        key = self._get_key_node(node)
        val = ("#ERROR#", node, node.parentCtx)
        self.vertices[key] = val
        kc = self._get_key_context(node.parentCtx)
        self.edges[kc, key] = 1

    def enterEveryRule(self, ctx):
        """
        event
        """
        key = self._get_key_context(ctx)
        self.vertices[key] = (key[-1], ctx, ctx.parentCtx)
        if ctx.parentCtx is not None:
            kc = self._get_key_context(ctx.parentCtx)
            self.edges[kc, key] = 1

        if self.verbose:  # pragma: no cover
            self.fLOG("+", type(ctx), ctx)
            self.fLOG("+", ctx.__dict__.keys())
            self.fLOG("  +", type(ctx.start), ctx.start)
            self.fLOG("  +", ctx.start.__dict__.keys())
            text = ctx.getText()
            print("  text:", text)
            if "node" in ctx.__dict__:
                self.fLOG("  +", type(ctx.node), ctx.node)
                self.fLOG("  +", ctx.node.__dict__.keys())

    def exitEveryRule(self, ctx):
        """
        event
        """
        if self.verbose:  # pragma: no cover
            self.fLOG("-", type(ctx), ctx)

    def to_networkx(self):
        """
        convert the graph into networkx

        @return     `networkx Graph <https://networkx.github.io/documentation/latest/tutorial/tutorial.html?highlight=graph>`_
        """
        import networkx  # pylint: disable=C0415
        verti = self.Vertices
        edges = self.Edges

        G = networkx.Graph()
        for v in verti:
            G.add_node(v)
        for k in edges:
            G.add_edge(k[0], k[1])
        return G

    def draw(self, ax=None):
        """
        draw the graph with networkx on matplotlib

        @param      matplotlib axis
        """
        if ax is None:
            import matplotlib.pyplot as plt  # pylint: disable=C0415
            fig, ax = plt.subplots()

        G = self.to_networkx()

        import networkx  # pylint: disable=C0415
        pos = networkx.spring_layout(G)  # positions for all nodes
        verti = self.Vertices
        labels = {p: verti[p][0] for p in pos}

        networkx.draw_networkx_nodes(G, pos, ax=ax)
        networkx.draw_networkx_edges(G, pos, ax=ax)
        networkx.draw_networkx_labels(G, pos, labels, font_size=16)

    def to_dot(self):
        """
        export the graph to DOT format

        @return         string
        """
        verti = self.Vertices
        edges = self.Edges

        ids = {}
        for k in verti:
            ids[k] = len(ids)

        rows = ["digraph {"]
        for k, v in verti.items():
            rows.append('%d [label="%s"];' % (
                ids[k], v[0].replace("\\", "\\\\").replace('"', '\\"')))
        for k in edges:
            a, b = ids[k[0]], ids[k[1]]
            rows.append("%d -> %d ;" % (a, b))
        rows.append("}")
        return "\n".join(rows)



"""
@file
@brief Shortcuts to datasource
"""
from .convert import dBase2df, dBase2sqllite
from .geodata import load_french_departements
from .http_retrieve import download_data, DownloadDataException



"""
@file
@brief Helpers to build grammars
This module requires `antlr4 <http://www.antlr.org/>`_.
and `antlr4-python3-runtime <https://pypi.python.org/pypi/antlr4-python3-runtime/>`_.
"""
import os
import sys
from pyquickhelper.loghelper import noLOG


def _is_syntax_is_missing(language):
    """
    Downloads the grammar for a specific language if
    the files is missing.

    @param      language        language: python, sqlite, ...
    @return                     grammar file
    """
    locations = {
        "R": "https://github.com/antlr/grammars-v4/tree/master/r/",
        "SQLite": "https://github.com/antlr/grammars-v4/blob/master/sqlite/",
        "CSharp": "https://github.com/antlr/grammars-v4/tree/master/csharp",
    }

    folder = os.path.dirname(__file__)
    filename = os.path.join(folder, language + ".g4")
    if os.path.exists(filename):
        return filename
    if language in locations:
        raise FileNotFoundError(
            "The grammar '{0}' is not available, you should get "
            "it from {1}".format(language, locations[language]))
    raise KeyError(
        "Unexpected language: '{0}', not in '{1}'".format(
            language, ",".join(locations.keys())))


def build_grammar(g4, version="4.10.1", fLOG=noLOG):
    """
    Compiles the grammar for a specific file.

    @param      g4          grammar format antlr4
    @param      version     version of *antlr4* to use, 4.9
    @param      fLOG        logging function
    @return                 list of files

    The compilation must be done with `antlr4 <http://www.antlr.org/>`_.
    It generates a lexer and a parser which can be imported in Python.
    The options for the command line are described at:
    `antlr4 options
    <https://theantlrguy.atlassian.net/wiki/display/ANTLR4/Options>`_.

    .. exref::
        :title: Builds a Antlr4 grammar

        See `grammars-v4 <https://github.com/antlr/grammars-v4>`_

        ::

            build_grammar("R.g4")
    """
    if not g4.endswith(".g4"):
        fold = os.path.abspath(os.path.dirname(__file__))
        g4 = os.path.join(fold, g4 + ".g4")

    url = "http://www.antlr.org/download/antlr-{0}-complete.jar".format(
        version)
    spl = url.split("/")
    domain, name = "/".join(spl[:-1]) + "/", spl[-1]
    folder = os.path.abspath(os.path.dirname(__file__))
    final = os.path.join(folder, name)

    if not os.path.exists(final):
        from ..datasource.http_retrieve import download_data
        name = download_data(name, website=domain, whereTo=folder)
        if not os.path.exists(name):
            raise FileNotFoundError("unable to download: " + url)

    path = os.environ.get("CLASSPATH", "")
    if name not in path:
        path = ".;{0}\\antlr-{1}-complete.jar".format(folder, version)
    else:
        path = ".;{0}\\antlr-{1}-complete.jar;{2}".format(
            folder,
            version,
            os.environ["CLASSPATH"])

    os.environ["CLASSPATH"] = path
    fLOG("CLASSPATH", os.environ["CLASSPATH"])

    # we remove -rc...
    version = version.split("-")[0]

    cmd = "org.antlr.v4.Tool "
    cmd += "-Dlanguage=Python3 "
    cmd += g4
    from pyquickhelper.loghelper import run_cmd
    out, err = run_cmd("java " + cmd, wait=True, fLOG=fLOG)

    def compiled():
        if "Lexer" in g4:
            lexer = g4.replace(".g4", ".tokens")
        else:
            lexer = g4.replace(".g4", ".py")
        return os.path.exists(lexer)

    if not compiled() or (len(err) > 0 and "error" in err):

        javapath = r'C:\Program Files\Java\jre7\bin\java.exe'
        os.environ["PATH"] = os.environ["PATH"] + ";" + javapath
        if sys.platform.startswith("win") and os.path.exists(javapath):
            out, err = run_cmd(
                '"' + javapath + '" ' + cmd, wait=True, fLOG=fLOG)
            if not compiled() or (len(err) > 0 and "error" in err):
                raise Exception(
                    "unable to compile: " +
                    final +
                    "\nCLASSPATH:\n" +
                    os.environ["CLASSPATH"] +
                    "\nERR:\n" +
                    err +
                    "\nCMD:\njava " +
                    cmd +
                    "\nYou should do it manually.")
        elif err:
            err_lines = err.split(err)
            err_lines = [_ for _ in err_lines if not _.startswith("warning(")]
            err2 = "\n".join(err_lines).strip("\n ")
            if len(err2) > 0:
                raise Exception(
                    "unable to compile: " +
                    final +
                    "\nCLASSPATH:\n" +
                    os.environ["CLASSPATH"] +
                    "\nERR:\n" +
                    err +
                    "\nCMD:\njava " +
                    cmd)

    if os.environ.get("USERNAME", os.environ.get("USER", "")) in g4:
        dest = os.path.dirname(g4)
        for name in os.listdir(dest):
            if "Parser" not in name and "Lexer" not in name and \
                    "Token" not in name and "Listener" not in name:
                continue
            full = os.path.join(dest, name)
            with open(full, "r", encoding="utf-8") as f:
                content = f.read()
            content1 = content.replace(dest, "")
            if content1 != content:
                fLOG("[build_grammar] modified", name)
                with open(full, "w", encoding="utf-8") as f:
                    f.write(content1)

    return out + "\n---ERR---\n" + err



# -*- coding: utf-8 -*-
"""
@file
@brief Complex joins with pandas.
"""
import pandas


def df_crossjoin(df1, df2, **kwargs):
    """
    Make a cross join (cartesian product) between two dataframes by using a constant temporary key.
    Also sets a MultiIndex which is the cartesian product of the indices of the input dataframes.
    Source: `Cross join / cartesian product between pandas DataFrames
    https://mkonrad.net/2016/04/16/cross-join--cartesian-product-between-pandas-dataframes.html>`_.

    @param      df1     dataframe 1
    @param      df2     dataframe 2
    @param      kwargs  keyword arguments that will be passed to pd.merge()
    @return             cross join of df1 and df2

    .. exref::
        :title: Cross join with a pandas dataframe

        .. runpython::
            :showcode:

            import pandas
            from pyensae.mlhelper import df_crossjoin
            df = pandas.DataFrame([{"x":3, "y": 4}, {"x":5, "y": 6}])
            jj = df_crossjoin(df, df.copy())

        A dataframe cannot be joined on itself, the second one musrt be copied.
    """
    df1['_tmpkey'] = 1
    df2['_tmpkey'] = 1
    res = pandas.merge(df1, df2, on='_tmpkey',
                       ** kwargs).drop('_tmpkey', axis=1)
    res.index = pandas.MultiIndex.from_product((df1.index, df2.index))
    df1.drop('_tmpkey', axis=1, inplace=True)
    df2.drop('_tmpkey', axis=1, inplace=True)
    return res



"""
@file
@brief dendogram, heatmap functionalities.

It comes from `linkage.py <https://raw.githubusercontent.com/biokit/biokit/master/biokit/viz/linkage.py>`_
which I copied here because the module does not properly work on Python 3 (import issues).
See also `biokit license <https://github.com/biokit/biokit/blob/master/LICENSE>`_.

:author: Thomas Cokelaer

"""


class Linkage:
    """
    Linkage used in other tools such as Heatmap,
    the class requires `scipy <http://www.scipy.org/>`_.
    """

    def __init__(self):
        """
        unused
        """
        pass

    def linkage(self, df, method, metric):
        """
        Mostly calls `linkage
        <https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html>`_.

        @param      df          dataframe
        @param      method      see `linkage
                                <https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html>`_
        @param      metric      see `linkage
                                <https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html>`_
        @return                 output of `linkage
                                <https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html>`_
        """
        from scipy.cluster.hierarchy import linkage
        from scipy.spatial.distance import pdist, squareform
        d = pdist(df)
        D = squareform(d)
        Y = linkage(D, method=method, metric=metric)
        return Y



# coding: latin-1
"""
@file

@brief @see cl Database
"""
from pyquickhelper.loghelper.flog import fLOG

module_odbc = None


class DatabaseObject:

    """
    Methods for database related to object, see @see cl Database.
    """

    def fill_table_with_objects(
            self, tablename, iterator_on, check_existence=False, skip_exception=False):
        """
        Fill a table of a database with object (from ``iterator_on``) following the interface:

        - a property ``schema_database`` which gives the schema
        - a property ``asrow`` which puts all values in a row
        - a property ``index`` which precises the index to unless it returns None
        - a property ``indexes`` which precises other indexes to create (optional)

        The property ``asrow`` must not include other objects, only their ids.
        If the table does not exists, it creates it.

        @param      tablename           name of a table (created if it does not exists)
        @param      iterator_on         iterator_on on StreamRSS object
        @param      check_existence     avoid adding an element if it already exists (based on the index columns)
        @param      skip_exception      skip exception while inserting an element

        The function do not check duplicate among elements sent in ``iterator_on``,
        it only checks duplicate between the new and the old ones (meaning in the database).
        """
        schema = None
        index = None
        indexes = None
        for _ in iterator_on:
            schema = _.schema_database
            index = _.index
            try:
                indexes = _.indexes
            except Exception:
                pass
            break

        if schema is None:
            # nothing to do, it is empty
            return

        if tablename not in self.get_table_list():
            fLOG("create table ", tablename)
            cursor = self.create_table(tablename, schema)
            if index is not None:
                self.create_index(
                    index + "_" + tablename + "_index",
                    tablename,
                            [index],
                    unique=True)
                if indexes is not None:
                    for ind in indexes:
                        if isinstance(ind, str):
                            self.create_index(
                                ind + "_" + tablename + "_index",
                                tablename,
                                [ind],
                                unique=False)
                        else:
                            self.create_index(
                                "_".join(ind) + "_" + tablename + "_index",
                                tablename,
                                ind,
                                unique=False)
            ite = map(lambda m: m.asrow, iterator_on)
            self.append_values(
                ite,
                tablename,
                schema,
                cursor=cursor,
                skip_exception=skip_exception)

        else:
            if check_existence:
                if index is None:
                    raise ValueError(
                        "unable to check existence because index property is not set up")

                def enumerate_nodup(iterator):
                    for it in iterator:
                        val = it.__dict__[index]
                        view = self.execute_view(
                            "SELECT * FROM %s WHERE %s=\"%s\";" %
                            (tablename, index, val))
                        if len(view) == 0:
                            yield it

                ite = map(lambda m: m.asrow, enumerate_nodup(iterator_on))
                self.append_values(
                    ite,
                    tablename,
                    schema,
                    cursor=None,
                    skip_exception=skip_exception)
            else:
                ite = map(lambda m: m.asrow, iterator_on)
                self.append_values(
                    ite,
                    tablename,
                    schema,
                    cursor=None,
                    skip_exception=skip_exception)

        self.commit()

    def enumerate_objects(self, table, classObject):
        """
        Iterator on objects assuming each row of a table is a object (classObject type).
        The classObject must have the following properties:
            - a staticmethod ``schema_database_read`` which gives the schema
            - the constructor must accept a constructor where parameter have the same name as the column names

        @param      table           table name
        @param      classObject     class object
        @return                     iterator on object

        Example:

        ::

        for blog in db.enumerate_objects ("blogs", StreamRSS):
            #...
        """
        schema = classObject.schema_database_read()
        for row in self.execute("SELECT * FROM %s" % table):
            di = {schema[i][0]: v for i, v in enumerate(row)}
            yield classObject(**di)



# Generated from \DOT.g4 by ANTLR 4.10.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .DOTParser import DOTParser
else:
    from DOTParser import DOTParser

# This class defines a complete listener for a parse tree produced by DOTParser.


class DOTListener(ParseTreeListener):

    # Enter a parse tree produced by DOTParser#graph.
    def enterGraph(self, ctx: DOTParser.GraphContext):
        pass

    # Exit a parse tree produced by DOTParser#graph.
    def exitGraph(self, ctx: DOTParser.GraphContext):
        pass

    # Enter a parse tree produced by DOTParser#stmt_list.
    def enterStmt_list(self, ctx: DOTParser.Stmt_listContext):
        pass

    # Exit a parse tree produced by DOTParser#stmt_list.
    def exitStmt_list(self, ctx: DOTParser.Stmt_listContext):
        pass

    # Enter a parse tree produced by DOTParser#stmt.
    def enterStmt(self, ctx: DOTParser.StmtContext):
        pass

    # Exit a parse tree produced by DOTParser#stmt.
    def exitStmt(self, ctx: DOTParser.StmtContext):
        pass

    # Enter a parse tree produced by DOTParser#attr_stmt.
    def enterAttr_stmt(self, ctx: DOTParser.Attr_stmtContext):
        pass

    # Exit a parse tree produced by DOTParser#attr_stmt.
    def exitAttr_stmt(self, ctx: DOTParser.Attr_stmtContext):
        pass

    # Enter a parse tree produced by DOTParser#attr_list.
    def enterAttr_list(self, ctx: DOTParser.Attr_listContext):
        pass

    # Exit a parse tree produced by DOTParser#attr_list.
    def exitAttr_list(self, ctx: DOTParser.Attr_listContext):
        pass

    # Enter a parse tree produced by DOTParser#a_list.
    def enterA_list(self, ctx: DOTParser.A_listContext):
        pass

    # Exit a parse tree produced by DOTParser#a_list.
    def exitA_list(self, ctx: DOTParser.A_listContext):
        pass

    # Enter a parse tree produced by DOTParser#edge_stmt.
    def enterEdge_stmt(self, ctx: DOTParser.Edge_stmtContext):
        pass

    # Exit a parse tree produced by DOTParser#edge_stmt.
    def exitEdge_stmt(self, ctx: DOTParser.Edge_stmtContext):
        pass

    # Enter a parse tree produced by DOTParser#edgeRHS.
    def enterEdgeRHS(self, ctx: DOTParser.EdgeRHSContext):
        pass

    # Exit a parse tree produced by DOTParser#edgeRHS.
    def exitEdgeRHS(self, ctx: DOTParser.EdgeRHSContext):
        pass

    # Enter a parse tree produced by DOTParser#edgeop.
    def enterEdgeop(self, ctx: DOTParser.EdgeopContext):
        pass

    # Exit a parse tree produced by DOTParser#edgeop.
    def exitEdgeop(self, ctx: DOTParser.EdgeopContext):
        pass

    # Enter a parse tree produced by DOTParser#node_stmt.
    def enterNode_stmt(self, ctx: DOTParser.Node_stmtContext):
        pass

    # Exit a parse tree produced by DOTParser#node_stmt.
    def exitNode_stmt(self, ctx: DOTParser.Node_stmtContext):
        pass

    # Enter a parse tree produced by DOTParser#node_id.
    def enterNode_id(self, ctx: DOTParser.Node_idContext):
        pass

    # Exit a parse tree produced by DOTParser#node_id.
    def exitNode_id(self, ctx: DOTParser.Node_idContext):
        pass

    # Enter a parse tree produced by DOTParser#port.
    def enterPort(self, ctx: DOTParser.PortContext):
        pass

    # Exit a parse tree produced by DOTParser#port.
    def exitPort(self, ctx: DOTParser.PortContext):
        pass

    # Enter a parse tree produced by DOTParser#subgraph.
    def enterSubgraph(self, ctx: DOTParser.SubgraphContext):
        pass

    # Exit a parse tree produced by DOTParser#subgraph.
    def exitSubgraph(self, ctx: DOTParser.SubgraphContext):
        pass

    # Enter a parse tree produced by DOTParser#id_.
    def enterId_(self, ctx: DOTParser.Id_Context):
        pass

    # Exit a parse tree produced by DOTParser#id_.
    def exitId_(self, ctx: DOTParser.Id_Context):
        pass


del DOTParser
